import 'dart:ffi' as ffi;

import 'package:equatable/equatable.dart';
import 'package:ffi/ffi.dart' as ffi;
import 'package:meta/meta.dart';
import 'package:more/more.dart';

import 'bindings.dart' as ffi;
import 'gio.dart';
import 'int8str.dart';
import 'settings_schema.dart';
import 'variant.dart';
import 'variant_type.dart';

/// High-level API for application settings
///
/// {@template gsettings}
/// The GSettings class provides a convenient API for storing and retrieving
/// application settings.
///
/// Reads and writes can be considered to be non-blocking. Reading settings
/// with GSettings is typically extremely fast: on approximately the same orde
/// of magnitude (but slower than) a hash-table lookup. Writing settings is also
/// extremely fast in terms of time to return to your application, but can be
/// extremely expensive for other threads and other processes. Many settings
/// backends (including dconf) have lazy initialisation which means in the
/// common case of the user using their computer without modifying any settings
/// a lot of work can be avoided. For dconf, the D-Bus service doesn't even need
/// to be started in this case. For this reason, you should only ever modify
/// GSettings keys in response to explicit user action. Particular care should
/// be paid to ensure that modifications are not made during startup -- for
/// example, when setting the initial value of preferences widgets.
///
/// ## Example
///
/// ```dart
/// import 'package:gsettings/gsettings.dart';
///
/// void main() {
///   final settings = GSettings(schemaId: 'org.gnome.desktop.interface');
///   final theme = settings.stringValue('gtk-theme');
///   print(theme); // e.g. 'Yaru'
///   settings.dispose();
/// }
/// ```
///
/// ## Schemas
///
/// When creating a GSettings instance, you have to specify a schema that
/// describes the keys in your settings and their types and default values, as
/// well as some other information.
///
/// Normally, a schema has a fixed path that determines where the settings are
/// stored in the conceptual global tree of settings. However, schemas can also
/// be relocatable, i.e. not equipped with a fixed path. This is useful e.g.
/// when the schema describes an 'account', and you want to be able to store an
/// arbitrary number of accounts.
///
/// Paths must start with and end with a forward slash character ('/') and must
/// not contain two sequential slash characters. Paths should be chosen based on
/// a domain name associated with the program or library to which the settings
/// belong. Examples of paths are `/org/gtk/settings/file-chooser/` and
/// `/ca/desrt/dconf-editor/`. Paths should not start with `/apps/`, `/desktop/`
/// or `/system/` as they often did in GConf.
///
/// Unlike other configuration systems (like GConf), GSettings does not restrict
/// keys to basic types like strings and numbers. GSettings stores values as
/// GVariant, and allows any GVariantType for keys. Key names are restricted to
/// lowercase characters, numbers and '-'. Furthermore, the names must begin
/// with a lowercase character, must not end with a '-', and must not contain
/// consecutive dashes.
///
/// Similar to GConf, the default values in GSettings schemas can be localized,
/// but the localized values are stored in gettext catalogs and looked up with
/// the domain that is specified in the `gettext-domain` attribute of the
/// `<schemalist>` or `<schema>` elements and the category that is specified in
/// the `l10n` attribute of the `<default>` element. The string which is
/// translated includes all text in the `<default>` element, including any
/// surrounding quotation marks.
///
/// The `l10n` attribute must be set to `messages` or `time`, and sets the
/// [locale category for
/// translation](https://www.gnu.org/software/gettext/manual/html_node/Aspects.html#index-locale-categories-1).
/// The `messages` category should be used by default; use `time` for
/// translatable date or time formats. A translation comment can be added as an
/// XML comment immediately above the `<default>` element — it is recommended to
/// add these comments to aid translators understand the meaning and
/// implications of the default value. An optional translation `context`
/// attribute can be set on the `<default>` element to disambiguate multiple
/// defaults which use the same string.
///
/// For example:
/// ```xml
///  <!-- Translators: A list of words which are not allowed to be typed, in
///       GVariant serialization syntax.
///       See: https://developer.gnome.org/glib/stable/gvariant-text.html -->
///  <default l10n='messages' context='Banned words'>['bad', 'words']</default>
/// ```
///
/// Translations of default values must remain syntactically valid serialized
/// GVariants (e.g. retaining any surrounding quotation marks) or runtime
/// errors will occur.
///
/// GSettings uses schemas in a compact binary form that is created
/// by the [glib-compile-schemas](https://developer.gnome.org/gio/stable/glib-compile-schemas.html)
/// utility. The input is a schema description in an XML format.
///
/// A DTD for the gschema XML format can be found here:
/// [gschema.dtd](https://gitlab.gnome.org/GNOME/glib/-/blob/master/gio/gschema.dtd)
///
/// The [glib-compile-schemas](https://developer.gnome.org/gio/stable/glib-compile-schemas.html)
/// tool expects schema files to have the extension `.gschema.xml`.
///
/// At runtime, schemas are identified by their ID (as specified in the
/// `id` attribute of the `<schema>` element). The convention for schema
/// ids is to use a dotted name, similar in style to a D-Bus bus name,
/// e.g. `org.gnome.SessionManager`. In particular, if the settings are
/// for a specific service that owns a D-Bus bus name, the D-Bus bus name
/// and schema IDshould match. For schemas which deal with settings not
/// associated with one named application, the ID should not use
/// StudlyCaps, e.g. `org.gnome.font-rendering`.
///
/// In addition to GVariant types, keys can have types that have enumerated
/// types. These can be described by a `<choice>`, `<enum>` or `<flags>`
/// element. The underlying type of such a key is string, but you can use
/// [enumValue] and [flagsValue] to access the numeric values corresponding to
/// the string value of enum and flags keys.
///
/// An example for default value:
/// ```xml
/// <schemalist>
///   <schema id="org.gtk.Test" path="/org/gtk/Test/" gettext-domain="test">
///
///     <key name="greeting" type="s">
///       <default l10n="messages">"Hello, earthlings"</default>
///       <summary>A greeting</summary>
///       <description>
///         Greeting of the invading martians
///       </description>
///     </key>
///
///     <key name="box" type="(ii)">
///       <default>(20,30)</default>
///     </key>
///
///     <key name="empty-string" type="s">
///       <default>""</default>
///       <summary>Empty strings have to be provided in GVariant form</summary>
///     </key>
///
///   </schema>
/// </schemalist>
/// ```
///
/// An example for ranges, choices and enumerated types:
/// ```xml
/// <schemalist>
///   <enum id="org.gtk.Test.myenum">
///     <value nick="first" value="1"/>
///     <value nick="second" value="2"/>
///   </enum>
///
///   <flags id="org.gtk.Test.myflags">
///     <value nick="flag1" value="1"/>
///     <value nick="flag2" value="2"/>
///     <value nick="flag3" value="4"/>
///   </flags>
///
///   <schema id="org.gtk.Test">
///
///     <key name="key-with-range" type="i">
///       <range min="1" max="100"/>
///       <default>10</default>
///     </key>
///
///     <key name="key-with-choices" type="s">
///       <choices>
///         <choice value='Elisabeth'/>
///         <choice value='Annabeth'/>
///         <choice value='Joe'/>
///       </choices>
///       <aliases>
///         <alias value='Anna' target='Annabeth'/>
///         <alias value='Beth' target='Elisabeth'/>
///       </aliases>
///       <default>'Joe'</default>
///     </key>
///
///     <key name='enumerated-key' enum='org.gtk.Test.myenum'>
///       <default>'first'</default>
///     </key>
///
///     <key name='flags-key' flags='org.gtk.Test.myflags'>
///       <default>["flag1","flag2"]</default>
///     </key>
///   </schema>
/// </schemalist>
/// ```
///
/// ## Vendor overrides
///
/// Default values are defined in the schemas that get installed by
/// an application. Sometimes, it is necessary for a vendor or distributor
/// to adjust these defaults. Since patching the XML source for the schema
/// is inconvenient and error-prone,
/// [glib-compile-schemas](https://developer.gnome.org/gio/stable/glib-compile-schemas.html)
/// reads so-called vendor override' files. These are keyfiles in the same
/// directory as the XML schema sources which can override default values. The
/// schema ID serves as the group name in the key file, and the values are
/// expected in serialized GVariant form, as in the following example:
/// ```
/// [org.gtk.Example]
/// key1='string'
/// key2=1.5
/// ```
///
/// glib-compile-schemas expects schema files to have the extension
/// `.gschema.override`.
///
/// ## Relocatable schemas
///
/// A relocatable schema is one with no `path` attribute specified on its
/// `<schema>` element. A relocatable GSettings object can be instantiated by
/// passing a path to GSettings constructor. Paths are typically constructed
/// dynamically from a constant prefix plus some form of instance identifier;
/// but they must still be valid GSettings paths. Paths could also be constant
/// and used with a globally installed schema originating from a dependency
/// library.
///
/// For example, a relocatable schema could be used to store geometry information
/// for different windows in an application. If the schema ID was
/// `org.foo.MyApp.Window`, it could be instantiated for paths
/// `/org/foo/MyApp/main/`, `/org/foo/MyApp/document-1/`,
/// `/org/foo/MyApp/document-2/`, etc. If any of the paths are well-known
/// they can be specified as <child> elements in the parent schema, e.g.:
/// ```xml
/// <schema id="org.foo.MyApp" path="/org/foo/MyApp/">
///   <child name="main" schema="org.foo.MyApp.Window"/>
/// </schema>
/// ```
/// {@endtemplate}
@immutable
class GSettings extends Equatable {
  final ffi.Pointer<ffi.GSettings> _ptr;

  /// Creates a new GSettings instance with the schema specified by [schemaId]
  /// and an optional [path].
  ///
  /// **Note:** Call [dispose] to release resources after you're done with the
  /// GSettings object.
  factory GSettings({required String schemaId, String? path}) {
    if (path == null) return GSettings._new(schemaId);
    return GSettings._newWithPath(schemaId, path);
  }

  /// @nodoc
  @internal
  const GSettings.fromPointer(this._ptr);

  /// @nodoc
  @internal
  ffi.Pointer<ffi.GSettings> toPointer() => _ptr;

  factory GSettings._new(String schema) {
    final cschema = schema.newInt8String();
    final settings = gio.g_settings_new(cschema);
    cschema.dispose();
    return GSettings.fromPointer(settings);
  }

  factory GSettings._newWithPath(String schema, String path) {
    final cschema = schema.newInt8String();
    final cpath = path.newInt8String();
    final gsettings = gio.g_settings_new_with_path(cschema, cpath);
    cschema.dispose();
    cpath.dispose();
    return GSettings.fromPointer(gsettings);
  }

  /// Ensures that all pending operations are complete.
  void sync() => gio.g_settings_sync();

  /// Releases all resources associated with this GSettings object.
  void dispose() => gio.g_object_unref(_ptr.cast());

  @override
  List<Object> get props => <Object>[schemaId, path];

  /// The ID of the schema that describes the types of keys for this GSettings
  /// object.
  String get schemaId => _getProperty('schema-id');

  /// The path within the backend where the settings are stored.
  String get path => _getProperty('path');

  String _getProperty(String name) {
    final value = ffi.calloc.allocate<ffi.GValue>(1);
    gio.g_value_init(value, ffi.G_TYPE_STRING);
    final property = name.newInt8String();
    gio.g_object_get_property(_ptr.cast(), property, value);
    final str = gio.g_value_get_string(value).toDartString();
    property.dispose();
    ffi.calloc.free(value);
    return str;
  }

  /// Gets the value that is stored in settings for [key].
  ///
  /// It is a programmer error to give a key that isn't contained in the schema
  /// for this GSettings object.
  Object? value(String key) {
    final ckey = key.newInt8String();
    final ffi.Pointer<ffi.GVariant> variant =
        gio.g_settings_get_value(_ptr, ckey);
    ckey.dispose();
    return variant.toObject();
  }

  String _typeName(String key) {
    final schema = GSettingsSchema.lookup(schemaId);
    final schemaKey = schema!.key(key);
    final typeName = schemaKey.typeName;
    schemaKey.dispose();
    return typeName;
  }

  /// Sets [key] in settings to [value].
  ///
  /// It is a programmer error to give a key that isn't contained in the schema
  /// for this GSttings object or for value to have the incorrect type, per the
  /// schema.
  bool setValue(String key, Object value) {
    final type = GVariantType.fromTypeName(_typeName(key));
    final cvar = value.toVariant(type).toPointer();
    final ckey = key.newInt8String();
    final res = gio.g_settings_set_value(_ptr, ckey, cvar);
    type.dispose();
    ckey.dispose();
    return res != 0;
  }

  /// Finds out if a [key] can be written or not.
  bool isWritable(String key) {
    final ckey = key.newInt8String();
    final res = gio.g_settings_is_writable(_ptr, ckey);
    ckey.dispose();
    return res != 0;
  }

  /// Changes the GSettings object into 'delay-apply' mode.
  ///
  /// In this mode, changes to settings are not immediately propagated to the
  /// backend, but kept locally until [apply] is called.
  void delay() {
    gio.g_settings_delay(_ptr);
  }

  /// Applies any changes that have been made to the settings.
  ///
  /// This function does nothing unless settings is in 'delay-apply' mode; see
  /// [delay]. In the normal case settings are always applied immediately.
  void apply() {
    gio.g_settings_apply(_ptr);
  }

  /// Reverts all non-applied changes to the settings.
  ///
  /// This function does nothing unless settings is in 'delay-apply' mode; see
  /// [delay]. In the normal case settings are always applied immediately.
  void revert() {
    gio.g_settings_revert(_ptr);
  }

  /// Returns whether the GSettings object has any unapplied changes.
  ///
  /// This can only be the case if it is in 'delayed-apply' mode; see [delay].
  bool hasUnapplied() {
    return gio.g_settings_get_has_unapplied(_ptr) != 0;
  }

  /// Creates a child settings object which has a base path of `base-path/[name]`,
  /// where base-path is the base path of settings .
  ///
  /// The schema for the child settings object must have been declared in the
  /// schema of settings using a <child> element.
  GSettings child(String name) {
    final cname = name.newInt8String();
    final gsettings = gio.g_settings_get_child(_ptr, cname);
    cname.dispose();
    return GSettings.fromPointer(gsettings);
  }

  /// Resets [key] to its default value.
  ///
  /// This call resets the key, as much as possible, to its default value. That
  /// might be the value specified in the schema or the one set by the
  /// administrator.
  void resetValue(String key) {
    final ckey = key.newInt8String();
    gio.g_settings_reset(_ptr, ckey);
    ckey.dispose();
  }

  /// Checks the "user value" of a [key], if there is one.
  ///
  /// The user value of a key is the last value that was set by the user.
  ///
  /// After calling [resetValue] this function should always return `null`
  /// (assuming something is not wrong with the system configuration).
  ///
  /// It is possible that [value] will return a different value than this
  /// function. This can happen in the case that the user set a value for a key
  /// that was subsequently locked down by the system administrator -- this
  /// function will return the user's old value.
  ///
  /// This function may be useful for adding a "reset" option to a UI or for
  /// providing indication that a particular value has been changed.
  ///
  // It is a programmer error to give a key that isn't contained in the schema
  // for this GSettings object.
  Object? userValue(String key) {
    final ckey = key.newInt8String();
    final cvariant = gio.g_settings_get_user_value(_ptr, ckey);
    final value = cvariant != ffi.nullptr ? cvariant.toObject() : null;
    ckey.dispose();
    return value;
  }

  /// Gets the "default value" of a [key].
  ///
  /// This is the value that would be read if [resetValue] were to be called on
  /// the [key].
  ///
  /// Note that this may be a different value than returned by
  /// [GSettingsSchemaKey.defaultValue] if the system administrator has provided
  /// a default value.
  ///
  /// Comparing the return values of [defaultValue] and [value] is not
  /// sufficient for determining if a value has been set because the user may
  /// have explicitly set the value to something that happens to be equal to the
  /// default. The difference here is that if the default changes in the future,
  /// the user's key will still be set.
  ///
  /// This function may be useful for adding an indication to a UI of what the
  /// default value was before the user set it.
  ///
  /// It is a programmer error to give a key that isn't contained in the schema
  /// for this GSettings object.
  Object? defaultValue(String key) {
    final ckey = key.newInt8String();
    final cvariant = gio.g_settings_get_default_value(_ptr, ckey);
    final value = cvariant.toObject();
    ckey.dispose();
    return value;
  }

  /// Gets the list of children on this GSettings object.
  ///
  /// The list is exactly the list of strings for which it is not an error to
  /// call [child].
  ///
  /// There is little reason to call this function from "normal" code, since you
  /// should already know what children are in your schema. This function may
  /// still be useful there for introspection reasons, however.
  List<String> children() {
    final list = <String>[];
    final cchildren = gio.g_settings_list_children(_ptr);
    var i = 0;
    while (cchildren[i] != ffi.nullptr) {
      list.add(cchildren[i].toDartString());
      ++i;
    }
    gio.g_strfreev(cchildren);
    return list;
  }

  /// Gets the boolean value that is stored in settings for [key].
  bool boolValue(String key) => value(key)! as bool;

  /// Gets the integer value that is stored in settings for [key].
  int intValue(String key) => value(key)! as int;

  /// Gets the double value that is stored in settings for [key].
  double doubleValue(String key) => value(key)! as double;

  /// Gets the string value that is stored in settings for [key].
  String stringValue(String key) => value(key)! as String;

  /// Gets the maybe value that is stored in settings for [key].
  Object? maybeValue(String key) => value(key);

  /// Gets the dict value that is stored in settings for [key].
  Map<Object, Object?> dictValue(String key) =>
      value(key)! as Map<Object, Object?>;

  /// Gets the boolean value that is stored in settings for [key].
  List<Object?> arrayValue(String key) => value(key)! as List<Object?>;

  /// Gets the byte-array value that is stored in settings for [key].
  List<int?> byteArrayValue(String key) => arrayValue(key).cast();

  /// Gets the string-array value that is stored in settings for [key].
  List<String?> stringArrayValue(String key) => arrayValue(key).cast();

  /// Gets the tuple value that is stored in settings for [key].
  Tuple tupleValue(String key) => value(key)! as Tuple;
}
