import 'dart:ffi' as ffi;

import 'package:equatable/equatable.dart';
import 'package:meta/meta.dart';
import 'package:more/more.dart';

import 'bindings.dart' as ffi;
import 'gio.dart';
import 'int8str.dart';
import 'variant.dart';
import 'variant_type.dart';

/// Introspection of settings schemas.
@immutable
class GSettingsSchema extends Equatable {
  final ffi.Pointer<ffi.GSettingsSchema> _ptr;

  /// Looks up a schema with the identifier [id].
  ///
  /// This function is not required for normal uses of GSettings but it may be
  /// useful to authors of plugin management systems or to those who want to
  /// introspect the content of schemas.
  ///
  /// If the schema isn't found, `null` is returned.
  static GSettingsSchema? lookup(String id) {
    final ptr = _lookupSchema(id);
    if (ptr == ffi.nullptr) return null;
    return GSettingsSchema.fromPointer(ptr);
  }

  /// @nodoc
  @internal
  const GSettingsSchema.fromPointer(this._ptr);

  /// @nodoc
  @internal
  ffi.Pointer<ffi.GSettingsSchema> toPointer() => _ptr;

  /// Releases all resources associated with this GSettingsSchema object.
  void dispose() => gio.g_settings_schema_unref(_ptr);

  @override
  List<Object> get props => <Object>[id, path ?? ''];

  static ffi.Pointer<ffi.GSettingsSchema> _lookupSchema(String id) {
    final cid = id.newInt8String();
    final source = gio.g_settings_schema_source_get_default();
    final schema = gio.g_settings_schema_source_lookup(source, cid, 1);
    cid.dispose();
    return schema;
  }

  /// Get the ID of the schema.
  String get id {
    return gio.g_settings_schema_get_id(_ptr).toDartString();
  }

  /// Gets the path associated with the schema, or `null`.
  ///
  /// Schemas may be single-instance or relocatable. Single-instance schemas
  /// correspond to exactly one set of keys in the backend database: those
  /// located at the path returned by this function.
  ///
  /// Relocatable schemas can be referenced by other schemas and can therefore
  /// describe multiple sets of keys at different locations. For relocatable
  /// schemas, this function will return 'null'.
  String? get path {
    final cstr = gio.g_settings_schema_get_path(_ptr);
    if (cstr == ffi.nullptr) return null;
    return cstr.toDartString();
  }

  /// Checks if schema has a key named [name].
  bool hasKey(String name) {
    final cname = name.newInt8String();
    final res = gio.g_settings_schema_has_key(_ptr, cname) != 0;
    cname.dispose();
    return res;
  }

  /// Gets the key named [name] from the schema.
  ///
  /// It is a programmer error to request a key that does not exist. See [keys].
  GSettingsSchemaKey key(String name) {
    final cname = name.newInt8String();
    final ckey = gio.g_settings_schema_get_key(_ptr, cname);
    cname.dispose();
    return GSettingsSchemaKey.fromPointer(ckey);
  }

  Iterable<String> _toStringList(ffi.Pointer<ffi.Pointer<ffi.Int8>> strv) {
    final list = <String>[];
    var i = 0;
    while (strv[i] != ffi.nullptr) {
      list.add(strv[i].toDartString());
      ++i;
    }
    gio.g_strfreev(strv);
    return list;
  }

  /// Gets the list of children in the schema.
  Iterable<String> get children {
    return _toStringList(gio.g_settings_schema_list_children(_ptr));
  }

  /// Introspects the list of keys on the schema.
  ///
  /// You should probably not be calling this function from "normal" code (since
  /// you should already know what keys are in your schema). This function is
  /// intended for introspection reasons.
  Iterable<String> get keys {
    return _toStringList(gio.g_settings_schema_list_keys(_ptr));
  }
}

/// Introspection of settings schema keys.
class GSettingsSchemaKey {
  final ffi.Pointer<ffi.GSettingsSchemaKey> _ptr;

  /// @nodoc
  @internal
  GSettingsSchemaKey.fromPointer(this._ptr) {
    gio.g_settings_schema_key_ref(_ptr);
  }

  /// @nodoc
  @internal
  ffi.Pointer<ffi.GSettingsSchemaKey> toPointer() => _ptr;

  /// Releases all resources associated with this GSettingsSchemaKey object.
  void dispose() => gio.g_settings_schema_key_unref(_ptr);

  /// Gets the name of the key.
  String get name => gio.g_settings_schema_key_get_name(_ptr).toDartString();

  /// Gets the summary for the key.
  ///
  /// If no summary has been provided in the schema, returns `null`.
  ///
  /// The summary is a short description of the purpose of the key; usually one
  /// short sentence. Summaries can be translated and the value returned from
  /// this function is is the current locale.
  ///
  /// This function is slow. The summary and description information for the
  /// schemas is not stored in the compiled schema database so this function has
  /// to parse all of the source XML files in the schema directory.
  String? get summary {
    final cstr = gio.g_settings_schema_key_get_summary(_ptr);
    if (cstr == ffi.nullptr) return null;
    return cstr.toDartString();
  }

  /// Gets the description for the key.
  ///
  /// If no description has been provided in the schema, returns `null`.
  ///
  /// The description can be one sentence to several paragraphs in length.
  /// Paragraphs are delimited with a double newline. Descriptions can be
  /// translated and the value returned from this function is is the current
  /// locale.
  ///
  /// This function is slow. The summary and description information for the
  /// schemas is not stored in the compiled schema database so this function has
  /// to parse all of the source XML files in the schema directory.
  String? get description {
    final cstr = gio.g_settings_schema_key_get_description(_ptr);
    if (cstr == ffi.nullptr) return null;
    return cstr.toDartString();
  }

  /// Gets the type name of the key.
  String get typeName {
    return gio.g_settings_schema_key_get_value_type(_ptr).typeName;
  }

  /// Gets the default value for the key.
  ///
  /// Note that this is the default value according to the schema. System
  /// administrator defaults and lockdown are not visible via this API.
  Object? get defaultValue {
    return gio.g_settings_schema_key_get_default_value(_ptr).toObject();
  }

  Tuple2<Object?, Object?>? get _range {
    final range = gio.g_settings_schema_key_get_range(_ptr);
    final tuple = range.toObject() as Tuple2<Object?, Object?>?;
    range.dispose();
    if (tuple == null || tuple.first != 'range') return null;
    return tuple.second as Tuple2<Object?, Object?>?;
  }

  /// Gets the minimum value for the key.
  Object? get minimumValue => _range?.first;

  /// Gets the maximum value for the key.
  Object? get maximumValue => _range?.second;

  /// Checks if the given value is of the correct type and within the permitted
  /// range for the key.
  ///
  /// It is a programmer error if value is not of the correct type -- you must
  /// check for this first.
  bool rangeCheck(Object value) {
    final type = GVariantType.fromTypeName(typeName);
    final variant = value.toVariant(type);
    final res =
        gio.g_settings_schema_key_range_check(_ptr, variant.toPointer()) != 0;
    type.dispose();
    variant.dispose();
    return res;
  }
}
