import 'dart:ffi' as ffi;

import 'package:ffi/ffi.dart' as ffi;
import 'package:more/more.dart';

import 'bindings.dart' as ffi;
import 'gio.dart';
import 'int8str.dart';
import 'variant_builder.dart';
import 'variant_type.dart';

class GVariant {
  final ffi.Pointer<ffi.GVariant> _ptr;

  GVariant.fromPointer(ffi.Pointer<ffi.GVariant> variant)
      : assert(variant != ffi.nullptr),
        _ptr = variant;

  ffi.Pointer<ffi.GVariant> toPointer() => _ptr;

  factory GVariant.fromType(GVariantType type) {
    final ctype = type.typeName.newInt8String();
    final ptr = gio.g_variant_new(ctype);
    ctype.dispose();
    return GVariant.fromPointer(ptr);
  }

  // ignore: avoid_positional_boolean_parameters
  GVariant.fromBoolean(bool value)
      : _ptr = gio.g_variant_new_boolean(value ? 1 : 0);
  GVariant.fromByte(int value) : _ptr = gio.g_variant_new_byte(value);
  GVariant.fromInt16(int value) : _ptr = gio.g_variant_new_int16(value);
  GVariant.fromUInt16(int value) : _ptr = gio.g_variant_new_uint16(value);
  GVariant.fromInt32(int value) : _ptr = gio.g_variant_new_int32(value);
  GVariant.fromUInt32(int value) : _ptr = gio.g_variant_new_uint32(value);
  GVariant.fromInt64(int value) : _ptr = gio.g_variant_new_int64(value);
  GVariant.fromUInt64(int value) : _ptr = gio.g_variant_new_uint64(value);
  GVariant.fromDouble(double value) : _ptr = gio.g_variant_new_double(value);
  GVariant.fromString(String value)
      : _ptr = gio.g_variant_new_take_string(value.newInt8String());

  GVariantType get type => _ptr.type;
  String get typeName => _ptr.typeName;

  bool toBoolean() => _ptr.toBoolean();
  int toByte() => _ptr.toByte();
  int toInt16() => _ptr.toInt16();
  int toUInt16() => _ptr.toUInt16();
  int toInt32() => _ptr.toInt32();
  int toUInt32() => _ptr.toUInt64();
  int toInt64() => _ptr.toInt64();
  int toUInt64() => _ptr.toUInt64();
  double toDouble() => _ptr.toDouble();
  String toDartString() => _ptr.toString();
  List<Object?> toList() => _ptr.toList();
  Map<Object, Object?> toMap() => _ptr.toMap();
  Tuple toTuple() => _ptr.toTuple();

  Object? toObject() => _ptr.toObject();

  void dispose() => _ptr.dispose();
}

extension GVariantPointer on ffi.Pointer<ffi.GVariant> {
  GVariantType get type =>
      GVariantType.fromPointer(gio.g_variant_get_type(this));
  String get typeName => gio.g_variant_get_type_string(this).toDartString();

  bool toBoolean() => gio.g_variant_get_boolean(this) != 0;
  int toByte() => gio.g_variant_get_byte(this);
  int toInt16() => gio.g_variant_get_int16(this);
  int toUInt16() => gio.g_variant_get_uint16(this);
  int toInt32() => gio.g_variant_get_int32(this);
  int toUInt32() => gio.g_variant_get_uint32(this);
  int toInt64() => gio.g_variant_get_int64(this);
  int toUInt64() => gio.g_variant_get_uint64(this);
  double toDouble() => gio.g_variant_get_double(this);

  String toDartString() {
    final ffi.Pointer<ffi.Int32> size = ffi.calloc<ffi.Int32>();
    final ffi.Pointer<ffi.Int8> cstr = gio.g_variant_get_string(this, size);
    final String string = cstr.toDartString(length: size.value);
    ffi.calloc.free(size);
    return string;
  }

  List<Object?> toList() {
    return List<Object?>.generate(_childCount, (int index) {
      return _childObject(index);
    });
  }

  MapEntry<Object, Object?> toMapEntry() {
    final key = _childObject(0);
    final value = _childObject(1);
    return MapEntry(key!, value);
  }

  Map<Object, Object?> toMap() {
    final map = <Object, Object?>{};
    for (var i = 0; i < _childCount; ++i) {
      final entry = _childObject(i) as MapEntry<Object, Object?>?;
      map[entry!.key] = entry.value;
    }
    return map;
  }

  Tuple toTuple() {
    switch (_childCount) {
      case 2:
        return Tuple2(_childObject(0), _childObject(1));
      case 3:
        return Tuple3(_childObject(0), _childObject(1), _childObject(2));
      case 4:
        return Tuple4(
          _childObject(0),
          _childObject(1),
          _childObject(2),
          _childObject(3),
        );
      case 5:
        return Tuple5(
          _childObject(0),
          _childObject(1),
          _childObject(2),
          _childObject(3),
          _childObject(4),
        );
      case 6:
        return Tuple6(
          _childObject(0),
          _childObject(1),
          _childObject(2),
          _childObject(3),
          _childObject(4),
          _childObject(5),
        );
      case 7:
        return Tuple7(
          _childObject(0),
          _childObject(1),
          _childObject(2),
          _childObject(3),
          _childObject(4),
          _childObject(5),
          _childObject(6),
        );
      case 8:
        return Tuple8(
          _childObject(0),
          _childObject(1),
          _childObject(2),
          _childObject(3),
          _childObject(4),
          _childObject(5),
          _childObject(6),
          _childObject(7),
        );
      case 9:
        return Tuple9(
          _childObject(0),
          _childObject(1),
          _childObject(2),
          _childObject(3),
          _childObject(4),
          _childObject(5),
          _childObject(6),
          _childObject(7),
          _childObject(8),
        );
      default:
        throw UnsupportedError('Tuple$_childCount (2-9)');
    }
  }

  Object? toMaybe() {
    final ffi.Pointer<ffi.GVariant> variant = gio.g_variant_get_maybe(this);
    if (variant == ffi.nullptr) return null;
    final maybe = variant.toObject();
    variant.dispose();
    return maybe;
  }

  Object? toVariant() {
    final ffi.Pointer<ffi.GVariant> variant = gio.g_variant_get_variant(this);
    final value = variant.toObject();
    variant.dispose();
    return value;
  }

  Object? toObject() {
    switch (type.typeName) {
      case GVariantTypeName.boolean:
        return toBoolean();
      case GVariantTypeName.byte:
        return toByte();
      case GVariantTypeName.int16:
        return toInt16();
      case GVariantTypeName.uint16:
        return toUInt16();
      case GVariantTypeName.int32:
        return toInt32();
      case GVariantTypeName.uint32:
        return toUInt32();
      case GVariantTypeName.int64:
        return toInt64();
      case GVariantTypeName.uint64:
        return toUInt64();
      case GVariantTypeName.double:
        return toDouble();
      case GVariantTypeName.string:
        return toDartString();
      default:
        if (type.isMaybe) return toMaybe();
        if (type.isDict) return toMap();
        if (type.isArray) return toList();
        if (type.isDictEntry) return toMapEntry();
        if (type.isTuple) return toTuple();
        if (type.isVariant) return toVariant();
        throw UnsupportedError('${type.typeName}.toObject()');
    }
  }

  int get _childCount => gio.g_variant_n_children(this);

  Object? _childObject(int index) {
    final ffi.Pointer<ffi.GVariant> variant =
        gio.g_variant_get_child_value(this, index);
    final child = variant.toObject();
    variant.dispose();
    return child;
  }

  void dispose() => gio.g_variant_unref(this);
}

extension GVariantObject on Object {
  GVariant toVariant(GVariantType type) {
    switch (type.typeName) {
      case GVariantTypeName.boolean:
        return GVariant.fromBoolean(this as bool);
      case GVariantTypeName.byte:
        return GVariant.fromByte(this as int);
      case GVariantTypeName.int16:
        return GVariant.fromInt16(this as int);
      case GVariantTypeName.uint16:
        return GVariant.fromUInt16(this as int);
      case GVariantTypeName.int32:
        return GVariant.fromInt32(this as int);
      case GVariantTypeName.uint32:
        return GVariant.fromUInt32(this as int);
      case GVariantTypeName.int64:
        return GVariant.fromInt64(this as int);
      case GVariantTypeName.uint64:
        return GVariant.fromUInt64(this as int);
      case GVariantTypeName.double:
        return GVariant.fromDouble(this as double);
      case GVariantTypeName.string:
        return GVariant.fromString(this as String);
      default:
        if (type.isMaybe) return toVariantMaybe(type);
        if (type.isDict) return toVariantDict(type);
        if (type.isArray) return toVariantArray(type);
        if (type.isDictEntry) return toVariantDictEntry(type);
        if (type.isTuple) return toVariantTuple(type);
        throw UnsupportedError('${type.typeName}.toVariant()');
    }
  }

  GVariant toVariantArray(GVariantType type) {
    return (this as List<Object>).toVariantArray(type);
  }

  GVariant toVariantDict(GVariantType type) {
    return (this as Map<Object, Object>).toVariantDict(type);
  }

  GVariant toVariantDictEntry(GVariantType type) {
    return (this as MapEntry<Object, Object>).toVariantDictEntry(type);
  }

  GVariant toVariantMaybe(GVariantType type) {
    return GVariantBuilder.build(type, toVariant(type.elementType));
  }

  GVariant toVariantTuple(GVariantType type) {
    return (this as Tuple).toVariantTuple(type);
  }
}

extension GVariantList on List<Object> {
  GVariant toVariantArray(GVariantType type) {
    return GVariantBuilder.buildAll(type, map((item) {
      return item.toVariant(type.elementType);
    }));
  }
}

extension GVariantDict on Map<Object, Object> {
  GVariant toVariantDict(GVariantType type) {
    return GVariantBuilder.buildAll(type, entries.map((entry) {
      return entry.toVariantDictEntry(type.elementType);
    }));
  }
}

extension GVariantDictEntry on MapEntry<Object, Object> {
  GVariant toVariantDictEntry(GVariantType type) {
    return GVariantBuilder.buildAll(type, <GVariant>[
      key.toVariant(type.keyType),
      value.toVariant(type.valueType)
    ]);
  }
}

extension GVariantTuple on Tuple {
  GVariant toVariantTuple(GVariantType type) {
    final it = type.itemTypes.iterator;
    return GVariantBuilder.buildAll(
      type,
      iterable.map((item) {
        return (item as Object).toVariant(it.moveNext()
            ? it.current
            : GVariantType.fromTypeName(GVariantTypeName.any));
      }),
    );
  }
}
