import 'dart:ffi' as ffi;

import 'bindings.dart' as ffi;
import 'gio.dart';
import 'int8str.dart';

class GVariantType {
  final ffi.Pointer<ffi.Int8> _cname;
  ffi.Pointer<ffi.GVariantType>? __ptr;
  GVariantType? __elementType;
  GVariantType? __keyType;
  GVariantType? __valueType;
  Iterable<GVariantType>? __itemTypes;

  GVariantType.fromTypeName(String typeName)
      : assert(GVariantTypeName.isValid(typeName)),
        _cname = typeName.newInt8String();

  GVariantType.fromPointer(ffi.Pointer<ffi.GVariantType> ptr)
      : assert(ptr != ffi.nullptr),
        _cname = ptr.typeName.newInt8String(),
        __ptr = ptr;

  ffi.Pointer<ffi.GVariantType> toPointer() => _ptr;

  void dispose() {
    _cname.dispose();
    __ptr?.dispose();
    __elementType?.dispose();
    __keyType?.dispose();
    __valueType?.dispose();
    // ignore: avoid_function_literals_in_foreach_calls
    __itemTypes?.forEach((type) => type.dispose());
  }

  ffi.Pointer<ffi.GVariantType> get _ptr {
    return __ptr ??= gio.g_variant_type_new(_cname);
  }

  String get typeName => _cname.toDartString();

  bool get isArray => _ptr.isArray;
  bool get isBasic => _ptr.isBasic;
  bool get isContainer => _ptr.isContainer;
  bool get isDefinite => _ptr.isDefinive;
  bool get isDict => isArray && elementType.isDictEntry;
  bool get isDictEntry => _ptr.isDictEntry;
  bool get isMaybe => _ptr.isMaybe;
  bool get isTuple => _ptr.isTuple;
  bool get isVariant => _ptr.isVariant;

  GVariantType get elementType {
    assert(isArray || isMaybe);
    return __elementType ??= GVariantType.fromTypeName(_ptr.elementTypeName);
  }

  GVariantType get keyType {
    assert(isDictEntry);
    return __keyType ??= GVariantType.fromTypeName(_ptr.keyTypeName);
  }

  GVariantType get valueType {
    assert(isDictEntry);
    return __valueType ??= GVariantType.fromTypeName(_ptr.valueTypeName);
  }

  Iterable<GVariantType> get itemTypes {
    assert(isTuple);
    return __itemTypes ??= _ptr.itemTypeNames.map((type) {
      return GVariantType.fromTypeName(type);
    });
  }
}

extension GVariantTypePointer on ffi.Pointer<ffi.GVariantType> {
  String get typeName {
    final length = gio.g_variant_type_get_string_length(this);
    final ctype = gio.g_variant_type_peek_string(this);
    return ctype.toDartString(length: length);
  }

  void dispose() => gio.g_variant_type_free(this);

  bool get isArray => gio.g_variant_type_is_array(this) != 0;
  bool get isBasic => gio.g_variant_type_is_basic(this) != 0;
  bool get isContainer => gio.g_variant_type_is_container(this) != 0;
  bool get isDefinive => gio.g_variant_type_is_definite(this) != 0;
  bool get isDictEntry => gio.g_variant_type_is_dict_entry(this) != 0;
  bool get isMaybe => gio.g_variant_type_is_maybe(this) != 0;
  bool get isTuple => gio.g_variant_type_is_tuple(this) != 0;
  bool get isVariant => gio.g_variant_type_is_variant(this) != 0;

  String get elementTypeName {
    assert(isArray || isMaybe);
    return gio.g_variant_type_element(this).typeName;
  }

  String get keyTypeName {
    assert(isDictEntry);
    return gio.g_variant_type_key(this).typeName;
  }

  String get valueTypeName {
    assert(isDictEntry);
    return gio.g_variant_type_value(this).typeName;
  }

  List<String> get itemTypeNames {
    assert(isTuple);
    final types = <String>[];
    var next = gio.g_variant_type_first(this);
    while (next != ffi.nullptr) {
      types.add(next.typeName);
      next = gio.g_variant_type_next(next);
    }
    return types;
  }
}

abstract class GVariantTypeName {
  const GVariantTypeName._();

  static bool isValid(String typeName) {
    final cname = typeName.newInt8String();
    final res = gio.g_variant_type_string_is_valid(cname);
    cname.dispose();
    return res != 0;
  }

  static const String boolean = 'b';
  static const String byte = 'y';
  static const String int16 = 'n';
  static const String uint16 = 'q';
  static const String int32 = 'i';
  static const String uint32 = 'u';
  static const String int64 = 'x';
  static const String uint64 = 't';
  static const String double = 'd';
  static const String string = 's';
  static const String objectPath = 'o';
  static const String signature = 'g';
  static const String variant = 'v';
  static const String handle = 'h';
  static const String unit = '()';
  static const String any = '*';
  static const String basic = '?';
  static const String maybe = 'm*';
  static const String array = 'a*';
  static const String tuple = 'r';
  static const String dictEntry = '{?*}';
  static const String dictionary = 'a{?*}';
  static const String stringArray = 'as';
  static const String objectPathArray = 'ao';
  static const String byteString = 'ay';
  static const String byteStringArray = 'aay';
  static const String varDict = 'a{sv}';
}
