import 'dart:ffi' as ffi;
import 'dart:io';

import 'package:gsettings/gsettings.dart';
import 'package:more/more.dart';
import 'package:test/test.dart';

import 'test_utils.dart';

void main() {
  setUp(() => setUpGSettings());

  test('instance', () {
    final settings1 = GSettings(schemaId: 'gsettings.dart.empty');
    final pointer1 = settings1.toPointer();
    expect(pointer1, isNot(equals(ffi.nullptr)));

    final settings2 = GSettings.fromPointer(pointer1);
    final pointer2 = settings2.toPointer();
    expect(pointer2, equals(pointer1));

    settings1.dispose();
    // ### TODO: settings2.dispose();
  });

  test('id & path', () {
    final types = GSettings(schemaId: 'gsettings.dart.types');
    expect(types.schemaId, equals('gsettings.dart.types'));
    expect(types.path, equals('/gsettings/dart/types/'));
    expect(types, equals(types));

    final same = GSettings(schemaId: 'gsettings.dart.types');
    expect(same.schemaId, equals('gsettings.dart.types'));
    expect(same.path, equals('/gsettings/dart/types/'));
    expect(same, equals(types));

    final none = GSettings(schemaId: 'gsettings.dart.none', path: '/none/');
    expect(none.schemaId, equals('gsettings.dart.none'));
    expect(none.path, equals('/none/'));
    expect(none, isNot(equals(same)));

    final empty = GSettings(schemaId: 'gsettings.dart.empty');
    expect(empty.schemaId, equals('gsettings.dart.empty'));
    expect(empty.path, equals('/gsettings/dart/empty/'));
    expect(empty, isNot(equals(same)));
    expect(empty, isNot(equals(none)));

    types.dispose();
    same.dispose();
    none.dispose();
    empty.dispose();
  });

  test('writable', () {
    final settings = GSettings(schemaId: 'gsettings.dart.empty');
    expect(settings.isWritable('bool-value'), isTrue);
    settings.dispose();
  });

  test('children', () {
    final parent = GSettings(schemaId: 'gsettings.dart.parent');
    expect(parent.children(), unorderedEquals(<String>['child1', 'child2']));

    final child1 = parent.child('child1');
    expect(child1.schemaId, 'gsettings.dart.none');
    child1.dispose();

    final child2 = parent.child('child2');
    expect(child2.schemaId, 'gsettings.dart.none');
    child2.dispose();

    parent.dispose();
  });

  test('delay', () {
    final settings = GSettings(schemaId: 'gsettings.dart.types');
    settings.delay();
    expect(settings.hasUnapplied(), isFalse);

    const key = 'bool-value';
    expect(settings.setValue(key, true), isTrue);
    expect(settings.hasUnapplied(), isTrue);

    settings.apply();
    expect(settings.hasUnapplied(), isFalse);

    expect(settings.setValue(key, true), isTrue);
    expect(settings.hasUnapplied(), isTrue);

    settings.revert();
    expect(settings.hasUnapplied(), isFalse);

    settings.dispose();
  });

  test('reset', () {
    final settings = GSettings(schemaId: 'gsettings.dart.types');

    const key = 'bool-value';
    final defaultValue = settings.defaultValue(key) as bool;

    expect(settings.setValue(key, !defaultValue), isTrue);
    expect(settings.value(key), equals(!defaultValue));
    expect(settings.userValue(key), equals(!defaultValue));

    settings.resetValue(key);
    expect(settings.value(key), equals(defaultValue));
    expect(settings.userValue(key), isNull);

    settings.dispose();
  });

  test('sync', () {
    final settings = GSettings(schemaId: 'gsettings.dart.types');
    final other = GSettings(schemaId: 'gsettings.dart.types');

    const key = 'bool-value';
    final defaultValue = settings.defaultValue(key) as bool;

    expect(settings.setValue(key, !defaultValue), isTrue);
    expect(settings.value(key), equals(!defaultValue));

    settings.sync();
    expect(other.value(key), equals(!defaultValue));

    settings.dispose();
    other.dispose();
  });

  group('types', () {
    late GSettings settings;

    setUp(() => settings = GSettings(schemaId: 'gsettings.dart.types'));
    tearDown(() => settings.dispose());

    test('bool', () {
      const key = 'bool-value';

      expect(settings.defaultValue(key), isTrue);
      expect(settings.setValue(key, false), isTrue);
      expect(settings.boolValue(key), isFalse);
    });

    test('byte', () {
      const key = 'byte-value';

      expect(settings.defaultValue(key), equals(255));
      expect(settings.setValue(key, 123), isTrue);
      expect(settings.intValue(key), equals(123));
    });

    test('int16', () {
      const key = 'int16-value';

      expect(settings.defaultValue(key), equals(32767));
      expect(settings.setValue(key, 123), isTrue);
      expect(settings.intValue(key), equals(123));
    });

    test('uint16', () {
      const key = 'uint16-value';

      expect(settings.defaultValue(key), equals(65535));
      expect(settings.setValue(key, 123), isTrue);
      expect(settings.intValue(key), equals(123));
    });

    test('int32', () {
      const key = 'int32-value';

      expect(settings.defaultValue(key), equals(2147483647));
      expect(settings.setValue(key, 123), isTrue);
      expect(settings.intValue(key), equals(123));
    });

    test('uint32', () {
      const key = 'uint32-value';

      // ### TODO: expect(settings.defaultValue(key), equals(4294967295));
      expect(settings.setValue(key, 123), isTrue);
      expect(settings.intValue(key), equals(123));
    });

    test('int64', () {
      const key = 'int64-value';

      // ### TODO: expect(settings.defaultValue(key), equals(9223372036854775807));
      expect(settings.setValue(key, 123), isTrue);
      expect(settings.intValue(key), equals(123));
    });

    test('double', () {
      const key = 'double-value';

      expect(settings.defaultValue(key), equals(123456.789));
      expect(settings.setValue(key, 123.456), isTrue);
      expect(settings.doubleValue(key), equals(123.456));
    });

    test('string', () {
      const key = 'string-value';

      expect(settings.defaultValue(key), equals('foo'));
      expect(settings.setValue(key, 'bar'), isTrue);
      expect(settings.stringValue(key), equals('bar'));
    });

    test('int-array', () {
      const key = 'array-int-values';
      const defaults = <int>[1, 2, 3];
      const ints = <int>[4, 5, 6];

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, ints), isTrue);
      expect(settings.arrayValue(key), equals(ints));
      expect(settings.byteArrayValue(key), equals(ints));
    });

    test('string-array', () {
      const key = 'array-string-values';
      const defaults = <String>['f', 'o', 'o'];
      const strings = <String>['foo', 'bar', 'baz'];

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, strings), isTrue);
      expect(settings.arrayValue(key), equals(strings));
      expect(settings.stringArrayValue(key), equals(strings));
    });

    test('int-int-dict', () {
      const key = 'dict-int-int-values';
      const defaults = <int, int>{1: 4, 2: 5, 3: 6};
      const entries = <int, int>{1: 7, 2: 8, 3: 9};

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.dictValue(key), equals(entries));
    });

    test('string-int-dict', () {
      const key = 'dict-string-int-values';
      const defaults = <String, int>{'a': 1, 'b': 2, 'c': 3};
      const entries = <String, int>{'a': 4, 'b': 5, 'c': 6};

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.dictValue(key), equals(entries));
    });

    test('string-string-dict', () {
      const key = 'dict-string-string-values';
      const defaults = <String, String>{'a': 'd', 'b': 'e', 'c': 'f'};
      const entries = <String, String>{'a': 'g', 'b': 'h', 'c': 'i'};

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.dictValue(key), equals(entries));
    });

    test('tuple2', () {
      const key = 'tuple2-values';
      const defaults = Tuple2<String, int>('a', 1);
      const entries = Tuple2<String, int>('b', 2);

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.tupleValue(key), equals(entries));
    });

    test('tuple3', () {
      const key = 'tuple3-values';
      const defaults = Tuple3<String, int, String>('a', 1, 'b');
      const entries = Tuple3<String, int, String>('b', 2, 'a');

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.tupleValue(key), equals(entries));
    });

    test('tuple4', () {
      const key = 'tuple4-values';
      const defaults = Tuple4<String, int, String, int>('a', 1, 'b', 2);
      const entries = Tuple4<String, int, String, int>('b', 2, 'a', 1);

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.tupleValue(key), equals(entries));
    });

    test('tuple5', () {
      const key = 'tuple5-values';
      const defaults =
          Tuple5<String, int, String, int, String>('a', 1, 'b', 2, 'c');
      const entries =
          Tuple5<String, int, String, int, String>('c', 2, 'b', 1, 'a');

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.tupleValue(key), equals(entries));
    });

    test('tuple6', () {
      const key = 'tuple6-values';
      const defaults =
          Tuple6<String, int, String, int, String, int>('a', 1, 'b', 2, 'c', 3);
      const entries =
          Tuple6<String, int, String, int, String, int>('c', 3, 'b', 2, 'a', 1);

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.tupleValue(key), equals(entries));
    });

    test('tuple7', () {
      const key = 'tuple7-values';
      const defaults = Tuple7<String, int, String, int, String, int, String>(
          'a', 1, 'b', 2, 'c', 3, 'd');
      const entries = Tuple7<String, int, String, int, String, int, String>(
          'd', 3, 'c', 2, 'b', 1, 'a');

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.tupleValue(key), equals(entries));
    });

    test('tuple8', () {
      const key = 'tuple8-values';
      const defaults =
          Tuple8<String, int, String, int, String, int, String, int>(
              'a', 1, 'b', 2, 'c', 3, 'd', 4);
      const entries =
          Tuple8<String, int, String, int, String, int, String, int>(
              'd', 4, 'c', 3, 'b', 2, 'a', 1);

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.tupleValue(key), equals(entries));
    });

    test('tuple9', () {
      const key = 'tuple9-values';
      const defaults =
          Tuple9<String, int, String, int, String, int, String, int, String>(
              'a', 1, 'b', 2, 'c', 3, 'd', 4, 'e');
      const entries =
          Tuple9<String, int, String, int, String, int, String, int, String>(
              'e', 4, 'd', 3, 'c', 2, 'b', 1, 'a');

      expect(settings.defaultValue(key), equals(defaults));
      expect(settings.setValue(key, entries), isTrue);
      expect(settings.tupleValue(key), equals(entries));
    });

    test('nullable-int', () {
      const key = 'nullable-int';

      expect(settings.defaultValue(key), isNull);
      expect(settings.setValue(key, 123), isTrue);
      expect(settings.intValue(key), equals(123));

      settings.resetValue(key);
      expect(settings.maybeValue(key), isNull);
    });

    test('nullable-string', () {
      const key = 'nullable-string';

      expect(settings.defaultValue(key), isNull);
      expect(settings.setValue(key, 'foo'), isTrue);
      expect(settings.stringValue(key), equals('foo'));

      settings.resetValue(key);
      expect(settings.maybeValue(key), isNull);
    });
  });
}
